/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2005/2/14 first release		*/
/*					*/
/****************************************/
#include <macro.h>
#include <string.h>
#include "mes.h"
#include "task.h"
#include "id.h"
#include <regdef.h>

typedef volatile struct {
	int	registers[REGNUM];
	int	command;
	int	*arg;
	int	stdio, std_err;
	int	svctime;
	int	ret_value;
	char	*cwd;
	char	*memory;
	char	**argv;
} TCB;

typedef volatile struct {
	TCB	*tcb;
	unsigned short flag;
} TCBPtr;

#define SYSTASK		2
#define P(n)		tcbptr->arg[n]
static TCB		*curtcb;
static volatile int	current_task, task_end;
static TCBPtr		*tcblist;
static TCB		*systcb;
static void		supertask(void);
static volatile int	counter;
static volatile int	ip_timeout;

#define UNKNOWN_FORMAT	0
#define ELF_FORMAT	1
#define COFF_FORMAT	2

extern char	master_devname[], shell_name[], stdio_device[];
void	svc_int(void);
void	time_int(void);
void	dummy_int(void);

#define task_sw() \
	if(current_task >= 0) {\
		memcpy((char*)(curtcb->registers), (char*)saveregs, sizeof(int) * REGNUM);\
	}\
	do {\
		if(++current_task > task_end) {\
			current_task = 0;\
		}\
	} while(tcblist[current_task].tcb == 0 || tcblist[current_task].flag & (TASKWAIT | TASKSTAY));\
	curtcb = tcblist[current_task].tcb;\
	memcpy((char*)saveregs, (char*)(curtcb->registers), sizeof(int) * REGNUM)

static int argsize(int argc, char** argv);
static void argcopy(int argc, char *dest, char **src);
static void interval_setup(void);
static int sleep(int id, int ms);
static void task_main(int argc, char **argv, int (*func)(int, char**));
static int attach(char *program, int start, char *tcb, int argc, char *argp, char *gottable, int stdfd);
static int __exit(int id, int value);
static int ping(int task_id, int ip, int size);
static int recvfrom(int task_id, int socket, char *data, int size, struct sockaddr *target);
static int tcp_connect(int task_id, int socket, struct sockaddr *target);
static int tcp_free(int task_id, int socket);
static int tcp_close(int task_id, int socket);
static int tcp_accept(int task_id, int socket, struct sockaddr *target);
static int tcp_read(int task_id, int socket, char *data, int size);
static void supertask(void);
static int ps_info(int pid, struct process *ps);
static int __kill(int pid);
static int cd(int task_id, char *_path);
static void set_task_flag(int id, int flag);
static void reset_task_flag(int id, int flag);
static int get_task_flag(int id);

static void* entries[SYSTASK] = {
	supertask, arp_main,
};

volatile int 	saveregs[REGNUM];
extern volatile char	timer_sw_flag;

static void set_task_flag(int id, int flag) {
	INT_DISABLE();
	tcblist[id].flag |= flag;
	INT_ENABLE();
}

static void reset_task_flag(int id, int flag) {
	INT_DISABLE();
	tcblist[id].flag &= ~flag;
	INT_ENABLE();
}

static int get_task_flag(int id) {
	int	flag;

	INT_DISABLE();
	flag = tcblist[id].flag;
	INT_ENABLE();
	return flag;
}

void set_task_stay(int id) {
	set_task_flag(id, TASKSTAY);
}

int task_time(void) {
	return counter;
}

int random(void) {
	unsigned int	data;

	data = counter ^ 0x12345678;
	data = ~data;
	return data;
}

int cur_task(void) {
	return current_task;
}

void task_sw_ctl(int flag) {
	if(flag == 0) {
		reset_task_flag(current_task, TASKNOSW);
	} else {
		set_task_flag(current_task, TASKNOSW);
	}
}

int set_arg(int _task_id, int cmd, int argc, int data) {
	TCB	*tcbptr;
	int	task_id;
	
	task_sw_ctl(1);
	if(_task_id == 0) {
		for(task_id = SYSTASK;task_id <= task_end;task_id++) {
			if(tcblist[task_id].tcb == 0 || get_task_flag(task_id) & TASKSTAY) continue;
			tcbptr = tcblist[task_id].tcb;
			if(P(7) == data) break;
		}
	} else {
		task_id = _task_id;
	}
	if(task_id >= SYSTASK && task_id <= task_end && tcblist[task_id].tcb != 0) {
		if(get_task_flag(task_id) & TASKWAIT) {
			tcbptr = tcblist[task_id].tcb;
			if(tcbptr->command == cmd) {
				INT_DISABLE();
				P(argc) = data;
				INT_ENABLE();
				task_sw_ctl(0);
				return 0;
			}
		}
	}
	task_sw_ctl(0);
	return -1;
}

int get_arg(int task_id, int cmd, int argc, int *data) {
	TCB	*tcbptr;

	task_sw_ctl(1);
	if(task_id >= SYSTASK && task_id <= task_end && tcblist[task_id].tcb != 0){
		if(get_task_flag(task_id) & TASKWAIT) {
			tcbptr = tcblist[task_id].tcb;
			if(tcbptr->command == cmd) {
				INT_DISABLE();
				*data = P(argc);
				INT_ENABLE();
				task_sw_ctl(0);
				return 0;
			}
		}
	}
	task_sw_ctl(0);
	return -1;
}

int get_reg(int task_id, int regnum, int *reg) {
	TCB	*tcbptr;

	if(regnum >= REGNUM) return -1;
	task_sw_ctl(1);
	if(task_id >= SYSTASK && task_id <= task_end && tcblist[task_id].tcb != 0) {
		tcbptr = tcblist[task_id].tcb;
		INT_DISABLE();
		*reg = tcbptr->registers[regnum];
		INT_ENABLE();
		task_sw_ctl(0);
		return 0;
	}
	task_sw_ctl(0);
	return -1;
}

static int argsize(int argc, char** argv) {
	int	i, size;

	size = 0;
	for(i = 0;i < argc;i++) size += strlen(argv[i]) + 1;
	size += sizeof(char*) * (argc + 1);
	return size;
}

static void argcopy(int argc, char *dest, char **src) {
	int	i, n, size;
	char	*ptr, **argv;

	argv = (char**)dest;
	ptr = &(dest[sizeof(char*) * (argc + 1)]);
	for(i = 0;i < argc;i++) {
		strcpy(ptr, src[i]);
		argv[i] = ptr;
		n = strlen(ptr);
		ptr = &(ptr[n + 1]);
	}
	argv[i] = 0;
}

/************************************/
/************ INTERRUPT *************/
/************************************/
void system_call(void) {
	if(saveregs[0] != 0) {
		tcblist[current_task].flag |= TASKWAIT;
		curtcb->command = saveregs[0];
		curtcb->arg = (int*)(saveregs[1]);
		curtcb->svctime = counter;
	}
	task_sw();
}

/************************************/
/************ INTERRUPT *************/
/************************************/
void interval(void) {
	int		i;
	volatile char	w;

	if((tcblist[current_task].flag & TASKNOSW) == 0) {
		task_sw();
	}
#if defined(__arm__)
	TC0_SR;
#endif
#if defined(__H8300H__)
	switch(timer_sw_flag) {
  #if defined(H8_3069)
	case TIMER0_SW:
		TISRA &= ~0x01;
		break;
	case TIMER1_SW:
		TISRA &= ~0x02;
		break;
	case TIMER2_SW:
		TISRA &= ~0x04;
		break;
  #endif
	default:
		TCSR_WR = 0xa500 | 0x1a;
		w = TCSR_RD;
		TCSR_WR = 0xa500 | 0x3a;
	}
#endif
#if defined(__sh2__)
	switch(timer_sw_flag) {
	case TIMER0_SW:
		TSR0 &= ~0x01;
		break;
	case TIMER1_SW:
		TSR1 &= ~0x01;
		break;
	case TIMER2_SW:
		TSR2 &= ~0x01;
		break;
	case TIMER3_SW:
		TSR3 &= ~0x01;
		break;
	case TIMER4_SW:
		TSR4 &= ~0x01;
		break;
	default:
		TCSR_WR = 0xa500 | 0x19;
		w = TCSR_RD;
		TCSR_WR = 0xa500 | 0x39;
	}
#endif
#if defined(__sh3__)
	WTCSR_WR = 0xa500 | 0x05;
	w = WTCSR;
	WTCSR_WR = 0xa500 | 0x85;
#endif
	counter++;
}

/************************************/
/************ INIT CODE *************/
/************************************/
static void interval_setup(void) {
#if defined(__arm__)
	set_handler(SuperTask, SWIRQ_ID, system_call);
	set_handler(SuperTask, TC0_ID, interval);
	TC0_CCR = TC_CLKDIS;
	TC0_CMR = TC_CAPT + TC_CLKS_MCK1024 + TC_CPCTRG
		 + TC_BURST_NONE + TC_ETRGEDG_EDGE_NONE;
	TCB0_BMR = TC_NONEXC0;
	TC0_CV = 0;
	TC0_RC = 33;
	TC0_IER = TC_CPCS;
	TC0_CCR = TC_CLKEN + TC_SWTRG;
#endif
#if defined(__H8300H__)
	set_handler(SuperTask, 8, svc_int);
	switch(timer_sw_flag) {
	case TIMER0_SW:
		IPRA &= ~0x04;
		IPRA |= 0x08;
  #if defined(H8_3069)
		set_handler(SuperTask, 24, time_int);
		TCR0 = 0x23;
		GRA0 = 2083;
		TISRA |= 0x10;
		TSTR |= 0x01;
  #endif
		break;
	case TIMER1_SW:
		IPRA &= ~0x02;
		IPRA |= 0x08;
  #if defined(H8_3069)
		set_handler(SuperTask, 28, time_int);
		TCR1 = 0x23;
		GRA1 = 2083;
		TISRA |= 0x20;
		TSTR |= 0x02;
  #endif
		break;
	case TIMER2_SW:
		IPRA &= ~0x01;
		IPRA |= 0x08;
  #if defined(H8_3069)
		set_handler(SuperTask, 32, time_int);
		TCR2 = 0x23;
		GRA2 = 2083;
		TISRA |= 0x40;
		TSTR |= 0x04;
  #endif
		break;
	default:
		set_handler(SuperTask, 20, time_int);
		TCSR_WR = 0xa500 | 0x1a;
		TCNT_WR = 0x5a00;
		TCSR_WR = 0xa500 | 0x3a;
	}
#endif
#if defined(__sh2__)
	set_handler(SuperTask, 32, svc_int);
	switch(timer_sw_flag) {
	case TIMER0_SW:
		set_handler(SuperTask, 88, time_int);
		TGR0A = 8192;
		TCR0 = 0x20;
		TSR0 &= ~0x01;
		TIER0 |= 0x01;
		TSTR |= 0x01;
		IPRD |= 0xf000;
		break;
	case TIMER1_SW:
		set_handler(SuperTask, 96, time_int);
		TGR1A = 8192;
		TCR1 = 0x20;
		TSR1 &= ~0x01;
		TIER1 |= 0x01;
		TSTR |= 0x02;
		IPRD |= 0x00f0;
		break;
	case TIMER2_SW:
		set_handler(SuperTask, 104, time_int);
		TGR2A = 8192;
		TCR2 = 0x20;
		TSR2 &= ~0x01;
		TIER2 |= 0x01;
		TSTR |= 0x04;
		IPRE |= 0xf000;
		break;
	case TIMER3_SW:
		set_handler(SuperTask, 112, time_int);
		TGR3A = 8192;
		TCR3 = 0x20;
		TSR3 &= ~0x01;
		TIER3 |= 0x01;
		TSTR |= 0x40;
		IPRE |= 0x00f0;
		break;
	case TIMER4_SW:
		set_handler(SuperTask, 120, time_int);
		TGR4A = 8192;
		TCR4 = 0x20;
		TSR4 &= ~0x01;
		TIER4 |= 0x01;
		TSTR |= 0x80;
		IPRF |= 0xf000;
		break;
	default:
		set_handler(SuperTask, 152, time_int);
		TCSR_WR = 0xa500 | 0x19;
		TCNT_WR = 0x5a00;
		TCSR_WR = 0xa500 | 0x39;
		IPRH |= 0xf000;
	}
#endif
#if defined(__sh3__)
	set_handler(SuperTask, 0x160, svc_int);
	set_handler(SuperTask, 0x560, time_int);
	WTCSR_WR = 0xa500 | 0x85;
#endif
	counter = 0;
}

void task_wakeup(int id, int value) {
	int	task_id;
	TCB	*tcbptr;

	for(task_id = SYSTASK;task_id <= task_end;task_id++) {
		if(tcblist[task_id].tcb == 0 || get_task_flag(task_id) & TASKSTAY) continue;
		if(get_task_flag(task_id) & TASKWAIT) {
			tcbptr = tcblist[task_id].tcb;
			if(tcbptr->command == _wait && P(0) == id) {
				reset_task_flag(task_id, TASKWAIT);
				tcbptr->ret_value = value;
			}
		}
	}
}

static int sleep(int id, int ms) {
	volatile unsigned int	diff, svctime;

	svctime = tcblist[id].tcb->svctime;
	diff = counter - svctime;
#if defined(__H8300H__)
	if(diff > ((ms / 3) << 2)) return 1;
#endif
#if defined(__sh2__)
	if(diff > ((ms / 3) << 4)) return 1;
#endif
#if defined(__sh3__)
	if(diff > (ms << 1)) return 1;
#endif
	return 0;
}

void wait_ms(int ms) {
	INT_DISABLE();
	tcblist[0].tcb->svctime = counter;
	INT_ENABLE();
	while(sleep(0, ms) == 0);
}

/************************************/
/************ INIT CODE *************/
/************************************/
void task_init(void) {
	static	char	curpath[32];
	int	*ptr, task_id, fd;
	char	*stackptr;

	tcblist = (TCBPtr *)__malloc(SuperTask, sizeof(TCBPtr) * MAXTASK);
	systcb = (TCB *)__malloc(SuperTask, sizeof(TCB) * SYSTASK);
	strcpy(curpath, "/ram0/");
	if(master_devname[0] != 0) {
		fd = OpenFile(SuperTask, master_devname, OptRead);
		if(fd != -1) {
			CloseFile(SuperTask, fd);
			strcpy(curpath, master_devname);
		}
	}
	ip_timeout = 1000;
	current_task = -1;
	task_end = SYSTASK - 1;
	bzero((char*)tcblist, sizeof(TCBPtr) * MAXTASK);	
	bzero((char*)systcb, sizeof(TCB) * SYSTASK);
	stackptr = fast_alloc(SuperTask, 0x1800);
	systcb[SVC_TASK].registers[SP_REG] = (int)&stackptr[0x1400 - 8];
	systcb[ARP_TASK].registers[SP_REG] = (int)&stackptr[0x400 - 8];
	for(task_id = 0;task_id < SYSTASK;task_id++) {
		tcblist[task_id].tcb = &systcb[task_id];
		tcblist[task_id].flag = 0;
		systcb[task_id].cwd = curpath;
		systcb[task_id].command = 0;
#if defined(__arm__)
		systcb[task_id].registers[SP_REG] = (int)entries[task_id] + 4;
		systcb[task_id].registers[16] = ARM_MODE_SYS;
#endif
#if defined(__H8300H__)
		ptr = (int*)(systcb[task_id].registers[SP_REG]);
		*ptr = (int)entries[task_id];
#endif
#if defined(__sh2__)
		ptr = (int*)(systcb[task_id].registers[SP_REG]);
		*ptr++ = (int)entries[task_id];			// PC
		*ptr++ = 0;					// SR
#endif
#if defined(__sh3__)
		systcb[task_id].registers[PC_REG] = (int)entries[task_id];
		systcb[task_id].registers[SR_REG] = 0x40000000;
#endif
	}
	interval_setup();
	int_enable();
	for(;;);
}

static void task_main(int argc, char **argv, int (*func)(int, char**)) {
	(*func)(argc, argv);
	for(;;);
}

static int attach(char *program, int start, char *tcb, int argc, char *argp, char *gottable, int stdfd) {
	int	new, *ptr;
	char	*path;
	TCB	*newtcb;

	newtcb = (TCB*)tcb;
	newtcb->memory = program;
	newtcb->command = 0;
	newtcb->stdio = (stdfd == 0) ? tcblist[0].tcb->stdio : stdfd;
	newtcb->std_err = tcblist[0].tcb->std_err;
	path = __malloc(SuperTask, MAX_PATH_LEN);
	if(path == 0) return -1;
	newtcb->ret_value = 0;
	newtcb->cwd = path;
	newtcb->argv = (char**)argp;
	strcpy((char*)newtcb->cwd, (char*)tcblist[0].tcb->cwd);
#if defined(__arm__)
	newtcb->registers[0] = argc;
	newtcb->registers[1] = (int)argp;
	newtcb->registers[2] = (int)&program[start];
	newtcb->registers[13] = (int)tcb - 4;
	newtcb->registers[SP_REG] = (int)task_main + 4;
	newtcb->registers[16] = ARM_MODE_USER;
#endif
#if defined(__H8300H__)
	newtcb->registers[0] = argc;
	newtcb->registers[1] = (int)argp;
	newtcb->registers[2] = (int)&program[start];
	newtcb->registers[5] = (int)gottable;
	newtcb->registers[SP_REG] = (int)tcb - 8;
	ptr = (int*)(newtcb->registers[7]);
	*ptr = (int)task_main;
#endif
#if defined(__sh2__)
	newtcb->registers[4] = argc;
	newtcb->registers[5] = (int)argp;
	newtcb->registers[6] = (int)&program[start];
	newtcb->registers[SP_REG] = (int)tcb - 8;
	ptr = (int*)(newtcb->registers[SP_REG]);
	*ptr++ = (int)task_main;		// PC
	*ptr++ = 0;				// SR
#endif
#if defined(__sh3__)
	newtcb->registers[4] = argc;
	newtcb->registers[5] = (int)argp;
	newtcb->registers[6] = (int)&program[start];
	newtcb->registers[SP_REG] = (int)tcb - 8;
	newtcb->registers[PC_REG] = (int)task_main;
	newtcb->registers[SR_REG] = 0x40000000;
#endif
	task_sw_ctl(1);
	for(new = SYSTASK;;new++) {
		if(new >= MAXTASK) {
			task_sw_ctl(0);
			return -1;
		}
		if(tcblist[new].tcb == 0) break;
	}
	if(new > task_end) task_end = new;
	tcblist[new].tcb = newtcb;
	tcblist[new].flag = 0;
	change_memid(new, newtcb->cwd);
	return new;
}

int get_stdio(int id) {
	int	task_id, stdio;

	task_id = (id == SuperTask) ? 0 : id;
	stdio = tcblist[task_id].tcb->stdio;
	return stdio;
}

int get_stderr(int id) {
	int	task_id, std_err;

	task_id = (id == SuperTask) ? 0 : id;
	std_err = tcblist[task_id].tcb->std_err;
	return std_err;
}

#define EXECBUFSIZ	128

int execute(int task_id, int argc, char **argv, int stdfd) {
	DirEnt	ent;
	char	name[EXECBUFSIZ], *coffarea, *program, *tcb, *argp, *chop, c, form;
	int	*gottable;
	int	start, size, stack, filesize, arg_size;
	int	i, fd, got, gotsiz, id, ret, addr, n;

	bzero(name, EXECBUFSIZ);
	if(argv[0][0] == '/') {
		strncpy(name, argv[0], EXECBUFSIZ - 16);
	} else {
		strncpy(name, (char*)tcblist[task_id].tcb->cwd, EXECBUFSIZ - 16);
		n = strlen(name);
		strncat(name, argv[0], EXECBUFSIZ - 2 - n);
	}

	chop = strrchr(name, '/');

	if(chop == NULL) return -1;

	c = chop[1];

	chop[1] = 0;

	filesize = 0;
	fd = SetDirent(SuperTask, name, &ent);

	if(fd == -1) return -1;

	chop[1] = c;
	while(NextRecord(SuperTask, fd, &ent) == 0) {
		if(strcmp(chop + 1, ent.name) == 0) {
			filesize = ent.length;
			break;
		}
	}
	__close(SuperTask, fd);
	*chop = '/';
	if(filesize == 0) return -1;
	form = UNKNOWN_FORMAT;
	fd = __open(SuperTask, name, OptRead);
	if(fd == -1) return -1;
	ret = elf_info(SuperTask, fd, filesize, &start, &size, &stack, &got, &gotsiz);
	if(ret == 0) {
		form = ELF_FORMAT;
	} else {
		coffarea = __malloc(SuperTask, filesize);
		if(coffarea == 0) {
			__close(SuperTask, fd);
			return -1;
		}
		__close(SuperTask, fd);
		fd = __open(SuperTask, name, OptRead);
		size = __read(SuperTask, fd, coffarea, filesize);
		size = get_program_size(coffarea);
		if(size == 0) {
			__close(SuperTask, fd);
			return -1;
		}
		form = COFF_FORMAT;
		stack = stack_size(coffarea);
		start = 0;
	}
	__close(SuperTask, fd);
	arg_size = argsize(argc, argv);
	program = fast_alloc(SuperTask, size + stack + sizeof(TCB) + arg_size + 6);
	if(program == 0) {
		if(form == COFF_FORMAT) __free(SuperTask, coffarea);
		return -1;
	}
	bzero(program, size + stack + sizeof(TCB) + 4);
	fd = __open(SuperTask, name, OptRead);
	if(fd == -1) {
		if(form == COFF_FORMAT) __free(SuperTask, coffarea);
		fast_free(SuperTask, program);
		return -1;
	}
	switch(form) {
	case ELF_FORMAT:
		ret = elf_load(SuperTask, fd, filesize, program);
		break;
	case COFF_FORMAT:
		ret = coff2bin(program, coffarea);
		__free(SuperTask, coffarea);
		break;
	}
	__close(SuperTask, fd);
	if(ret== -1) {
		fast_free(SuperTask, program);
		return -1;
	}
	addr = (int)program + size + stack + 3;
	addr >>= 2;
	addr <<= 2;
	tcb = (char*)addr;
	addr += sizeof(TCB) + 3;
	addr >>= 2;
	addr <<= 2;
	argp = (char*)addr;
	argcopy(argc, argp, argv);
	gottable = (int*)0;
	if(form == ELF_FORMAT) {
		gottable = (int*)&program[got];
		for(i = 0;i < (gotsiz / sizeof(int));i++) {
			if(gottable[i] != 0) {
				gottable[i] += (int)program;
			}
		}
	}
	id = attach(program, start, tcb, argc, argp, (char*)gottable, stdfd);
	fast_change_memid(id, program);
	strcpy((char*)tcblist[id].tcb->cwd, (char*)tcblist[task_id].tcb->cwd);
	task_sw_ctl(0);
	return id;
}

static int __exit(int id, int value) {
	task_wakeup(id, value);
	if((get_task_flag(id) & TASKSTAY) == 0) {
		free_device(id);
		free_file(id);
		udp_clean(id);
		tcp_clean(id);
		free_memory(id);
		tcblist[id].tcb = 0;
		tcblist[id].flag = 0;
	}
	return 0;
}

static int ping(int task_id, int ip, int size) {
	TCB		*tcbptr;
	int		seq;
	volatile int	diff, svctime;

	tcbptr = tcblist[task_id].tcb;
	if(P(7) == 0) {
		seq = icmp_send(task_id, ip, size);
		if(seq == -1) {
			return -1;
		}
		INT_DISABLE();
		P(7) = seq;
		P(6) = 0;
		INT_ENABLE();
	}
	if(P(6) != 0) {
		if(P(6) == P(7)) return 1;
		return -1;
	}
	svctime = tcblist[task_id].tcb->svctime;
	diff = counter - svctime;
#if defined(__H8300H__)
	if(diff > 5000) return -1;
#endif
#if defined(__sh2__)
	if(diff > 40000) return -1;
#endif
#if defined(__sh3__)
	if(diff > 2000) return -1;
#endif
	return -2;
}

static int recvfrom(int task_id, int socket, char *data, int size, struct sockaddr *target) {
	TCB		*tcbptr;
	int		_task_id;
	volatile int	diff, svctime, timeout;
	struct sockaddr	bind;

	tcbptr = tcblist[task_id].tcb;
	_task_id = get_udp_bind(socket, &bind);
	if(_task_id != task_id) return -1;
	if(udp_read(task_id, socket, data, size, target) != -1) return P(5);
	if(P(7) == 0) {
		INT_DISABLE();
		P(7) = bind.sin_port;
		P(6) = 0;
		INT_ENABLE();
	}
	if(P(6) != 0) {
		if(P(6) == P(7)) return P(5);
		return -1;
	}
	svctime = tcblist[task_id].tcb->svctime;
	diff = counter - svctime;
	timeout = get_udp_timeout(task_id, socket);
	if(diff >  timeout) return -1;
	return -2;
}

static int tcp_connect(int task_id, int socket, struct sockaddr *target) {
	TCB	*tcbptr;
	int	ret;

	tcbptr = tcblist[task_id].tcb;
	if(P(7) == 0) {
		ret = tcp_synsend(task_id, socket, target);
		if(ret == -1) return -1;
		INT_DISABLE();
		P(7) = task_id;
		P(6) = 0;
		INT_ENABLE();
	}
	if(P(6) != 0) {
		if(P(6) == P(7)) return 0;
		return -1;
	}
	if(check_socket(socket, task_id) == -1) return -1;
	return -2;
}

static int tcp_free(int task_id, int socket) {
	TCB	*tcbptr;
	int	ret;

	tcbptr = tcblist[task_id].tcb;
	if(P(7) == 0) {
		ret = tcp_reset_socket(task_id, socket);
		if(ret == -1) return -1;
		INT_DISABLE();
		P(7) = task_id;
		INT_ENABLE();
	}
	if(P(6) != 0) {
		if(P(6) == P(7)) return 0;
		return -1;
	}
	if(check_socket(socket, task_id) == -1) return -1;
	return -2;
}

static int tcp_close(int task_id, int socket) {
	TCB	*tcbptr;
	int	ret;

	tcbptr = tcblist[task_id].tcb;
	if(P(7) == 0) {
		ret = tcp_close_socket(task_id, socket);
		if(ret == -1) return -1;
		INT_DISABLE();
		P(7) = task_id;
		INT_ENABLE();
	}
	if(P(6) != 0) {
		if(P(6) == P(7)) return 0;
		return -1;
	}
	if(check_socket(socket, task_id) == -1) return -1;
	return -2;
}

static int tcp_accept(int task_id, int socket, struct sockaddr *target) {
	TCB	*tcbptr;
	int	ret, a;

	tcbptr = tcblist[task_id].tcb;
	if(P(7) == 0) {
		ret = tcp_synrecv(task_id, socket, target);
		INT_DISABLE();
		P(7) = task_id;
		P(6) = 0;
		INT_ENABLE();
		return ret;
	}
	if(P(6) != 0) {
		if(P(6) == P(7)) return P(5);
		return -1;
	}
	if(check_socket(socket, task_id) == -1) return -1;
	return -2;
}

static int tcp_read(int task_id, int socket, char *data, int size) {
	TCB		*tcbptr;
	int		ret;
	volatile int	diff, svctime, timeout;

	tcbptr = tcblist[task_id].tcb;
	ret = tcp_get(task_id, socket, data, size);
	if(ret == -1) return -1;
	if(ret > 0) return ret;
	svctime = tcblist[task_id].tcb->svctime;
	diff = counter - svctime;
	timeout = get_tcp_timeout(task_id, socket);
	if(diff > timeout) return -1;
	return -2;
}

char*	__cwd(int task_id) {
	TCB	*tcbptr;

	tcbptr = tcblist[task_id].tcb;
	return tcbptr->cwd;
}

static int	cd(int task_id, char *_path) {
	int	fd, max, i;
	TCB	*tcbptr;
	char	*path, *src, *dst;
	
	tcbptr = tcblist[task_id].tcb;
	max = strlen(_path);
	if(_path[0] != '/') max += strlen(tcbptr->cwd);
	if(max >= MAX_PATH_LEN - 2) return -1;
	path = __malloc(SuperTask, MAX_PATH_LEN);
	if(path == 0) return -1;
	path[0] = 0;
	if(_path[0] != '/') strcat(path, tcbptr->cwd);
	strcat(path, _path);
	if(path[strlen(path) - 1] != '/') strcat(path, "/");
	fd = SetDirent(SuperTask, path, 0);
	if(fd == -1) {
		__free(SuperTask, path);
		return -1;
	}
	__close(SuperTask, fd);
	dst = tcbptr->cwd;
	src = path;
	*dst++ = *src++;
	for(;;){
		while(*src != '/') *dst++ = *src++;		
		*dst++ = *src++;
		while(memcmp(src, "../",3) == 0) {
			--dst;
			while(*--dst != '/');
			dst++;
			src = &src[3];
		}
		if(*src == 0) {
			*dst++ = *src++;
			break;
		}
	}
	__free(SuperTask, path);
	return 0;
}

static int ps_info(int pid, struct process *ps) {
	TCB	*tcbptr;
	int	*ptr;

	if(pid < SYSTASK) return 0;
	if(pid > task_end) return -1;
	if(tcblist[pid].tcb == 0) return 0;
	tcbptr = tcblist[pid].tcb;
	ps->flag = (get_task_flag(pid) & (TASKWAIT | TASKSTAY)) ? PS_WAIT : 0;
	ps->cwd = tcbptr->cwd;
	ps->argv = tcbptr->argv;
	ps->address = (int)tcbptr->memory;
	ps->stack = (int)tcbptr->registers[SP_REG];
	ps->tcb = (int)tcblist[pid].tcb;
#if defined(__sh3__)
	ps->pc = (int)tcbptr->registers[PC_REG];
#endif
#if defined(__H8300H__) || defined(__sh2__)
	ptr = (int*)tcbptr->registers[SP_REG];
	ps->pc = *ptr;
#endif
	return pid;
}

static int __kill(int pid) {
	if(pid < SYSTASK) return -1;
	if(pid > task_end) return -1;
	if(tcblist[pid].tcb == 0) return -1;
	if(get_task_flag(pid) & TASKSTAY) {
		if(del_device(pid) == -1) return -1;
		reset_task_flag(pid, TASKSTAY);
	}
	__exit(pid, -1);
	return 0;
}

#define INIBUFSIZ	256
#define INIARGNUM	32

void init_process(void) {
	int	stdio, task_id, retval, pid, fd, n, i, argn;
	char	*arg[INIARGNUM], line[INIBUFSIZ];

	fd = command_trap(_open_, "module.sys", OptRead);
	if(fd != -1) {
		do {
			bzero(line, INIBUFSIZ);
			for(n = 0;n < (INIBUFSIZ - 1);n++) {
				retval = command_trap(_read_, fd, &line[n], 1);
				if(retval == 0) break;
				if(line[n] < ' ') break;
			}
			line[n] = 0;
			if(memcmp(line, "end", 3) == 0) break;
			i = argn = 0;
			arg[argn++] = &(line[i]);
			while(line[i] != 0) {
				if(line[i] == ' ') {
					line[i] = 0;
					if(argn >= (INIARGNUM - 1)) break;
					arg[argn++] = &(line[i + 1]);
				}
				i++;
			}
			arg[argn] = 0;
			command_trap(_execute, argn, arg, 0);
		} while(retval > 0);
		command_trap(_close_, fd);
	}

	if(stdio_device[0] == 0) {
#if defined(__arm__)
		strcpy(stdio_device, "com0");
#endif
#if defined(__H8300H__) || defined(__sh2__)
		strcpy(stdio_device, "com1");
#endif
#if defined(__sh3__)
		strcpy(stdio_device, "com2");
#endif
	}
	while((stdio = command_trap(_open_, stdio_device, 0)) == -1);
	for(task_id = 0;task_id < SYSTASK;task_id++) {
		tcblist[task_id].tcb->stdio = stdio;
		tcblist[task_id].tcb->std_err = stdio;
	}

#if defined(__sh__)
	arg[0] = "shell.exe";
#endif
#if defined(__H8300H__) || defined(__arm__)
	arg[0] = "shell.elf";
#endif
	if(shell_name[0] != 0) arg[0] = shell_name;
	arg[1] = 0;
	retval = command_trap(_execute, 1, arg, 0);
	if(retval == -1) {
		command_trap(_write_, 0, arg[0], strlen(arg[0]));
		command_trap(_write_, 0, " not found\r", 11);
	}
	arg[0] = "hdd0";
	arg[1] = 0;
	retval = command_trap(_mount, 1, arg, 0);
}

static void supertask(void) {
	int	task_id, complete, retval, cnt, chip;
	int	id;
	char	c;
	char	*arg[2];
	TCB	*tcbptr;

	cnt = 0;
	for(;;) {
		ip_interval();
		device_interval();
		for(task_id = 1;task_id <= task_end;task_id++) {
			if(tcblist[task_id].tcb == 0 || get_task_flag(task_id) & TASKSTAY) continue;
			id = task_id;
			if(task_id < SYSTASK) id = SuperTask;
			if(get_task_flag(task_id) & TASKWAIT) {
				tcbptr = tcblist[task_id].tcb;
				task_sw_ctl(0);
				complete = 1;
				switch(tcbptr->command) {
				case _sysmalloc_:
					retval = (int)__malloc(P(0), P(1));
					break;
				case _sysfree_:
					retval = __free(P(0), (char*)P(1));
					break;
				case _open_:
					retval = __open(id, (char*)P(0), P(1));
					break;
				case _close_:
					retval = __close(id, P(0));
					break;
				case _read_:
					retval = __read(id, P(0), (char*)P(1), P(2));
					if(retval == -2) complete = 0;
					break;
				case _write_:
					retval = __write(id, P(0), (char*)P(1), P(2));
					if(retval == -2) complete = 0;
					break;
				case _ioctl_:
					retval = __ioctl(id, P(0), P(1), P(2));
					break;
				case _seek_:
					retval = __seek(id, P(0), P(1));
					break;
				case _get_pid:
					retval = task_id;
					break;
				case _int_enable:
					reset_task_flag(id, TASKNOSW);
					break;
				case _int_disable:
					set_task_flag(id, TASKNOSW);
					break;
				case _set_dirent:
					retval = SetDirent(id, (char*)P(0), (DirEnt*)P(1));
					break;
				case _next_record:
					retval = NextRecord(id, P(0), (DirEnt*)P(1));
					break;
				case _delete:
					retval = DeleteFile(id, (char*)P(0));
					break;
				case _set_handler:
					retval = set_handler(id, P(0), (void*)P(1));
					break;
				case _execute:
					retval = execute(id, P(0), (char**)P(1), P(2));
					break;
				case _wait:
					complete = 0;
					break;
				case _exit:
					retval = __exit(id, P(0));
					break;
				case _cd:
					retval = cd(id, (char*)P(0));
					break;
				case _cwd:
					retval = (int)__cwd(id);
					break;
				case _malloc_:
					retval = (int)__malloc(id, P(0));
					break;
				case _free_:
					retval = __free(id, (char*)P(0));
					break;
				case _sleep:
					retval = sleep(id, P(0));
					if(retval == 0) complete = 0;
					break;
				case _ifconfig:
					retval = ifconfig((char*)P(0), P(1), P(2));
					break;
				case _ifdown:
					retval = ifdown(P(0));
					break;
				case _net_ioctl_:
					retval = net_ioctl(P(0), P(1), P(2));
					break;
				case _set_gateway:
					retval = set_gateway(P(0));
					break;
				case _ping:
					retval = ping(id, P(0), P(1));
					if(retval == -2) complete = 0;
					break;
				case _udp_socket:
					retval = (int)udp_socket(id);
					break;
				case _udp_free:
					retval = udp_free(id, P(0));
					break;
				case _udp_bind:
					retval = udp_bind(id, P(0), (struct sockaddr*)P(1));
					break;
				case _sendto:
					retval = sendto(id, P(0), (char*)P(1), P(2), (struct sockaddr*)P(3));
					break;
				case _recvfrom:
					retval = recvfrom(id, P(0), (char*)P(1), P(2), (struct sockaddr*)P(3));
					if(retval == -2) complete = 0;
					break;
				case _get_ipinfo:
					retval = get_ipinfo(P(0), (int*)P(1), (int*)P(2), (char*)P(3));
					break;
				case _udp_timeout:
					retval = udp_timeout(id, P(0), P(1));
					break;
				case _tcp_socket:
					retval = tcp_socket(id, P(0));
					break;
				case _tcp_free:
					retval = tcp_free(id, P(0));
					if(retval == -2) complete = 0;
					break;
				case _tcp_bind:
					retval = tcp_bind(id, P(0), (struct sockaddr*)P(1));
					break;
				case _tcp_connect:
					retval = tcp_connect(id, P(0), (struct sockaddr*)P(1));
					if(retval == -2) complete = 0;
					break;
				case _tcp_listen:
					retval = tcp_listen(id, P(0), P(1));
					break;
				case _tcp_accept:
					retval = tcp_accept(id, P(0), (struct sockaddr*)P(1));
					if(retval == -2) complete = 0;
					break;
				case _tcp_read_:
					retval = tcp_read(id, P(0), (char*)P(1), P(2));
					if(retval == -2) complete = 0;
					break;
				case _tcp_write_:
					retval = tcp_write(id, P(0), (char*)P(1), P(2));
					break;
				case _tcp_timeout:
					retval = tcp_timeout(id, P(0), P(1));
					break;
				case _shm_get:
					retval = shm_get(P(0), P(1));
					break;
				case _shm_at:
					retval = (int)shm_at(P(0));
					break;
				case _shm_dt:
					retval = shm_dt(P(0));
					break;
				case _mount:
					retval = mount((char*)P(0));
					break;
				case _eject:
					retval = eject((char*)P(0));
					break;
				case _format:
					retval = format((char*)P(0));
					break;
				case _setip:
					retval = setip(P(0), P(1), P(2));
					break;
				case _get_filesize:
					retval = GetFileSize(id, P(0));
					break;
				case _tcp_change_taskid:
					retval = tcp_change_taskid(id, P(0));
					break;
				case _tcp_close:
					retval = tcp_close(id, P(0));
					if(retval == -2) complete = 0;
					break;
				case _dns_server:
					retval = dns_server(P(0));
					break;
				case _get_netnum:
					retval = get_netnum((char*)P(0));
					break;
				case _dhcp_server:
					retval = dhcp_server(P(0), P(1));
					break;
				case _random:
					retval = (int)random();
					break;
				case _usb_mount:
					retval = usb_mount(P(0));
					break;
				case _usb_eject:
					retval = usb_eject(P(0));
					break;
				case _ps_info:
					retval = ps_info(P(0), (struct process*)P(1));
					break;
				case _kill_:
					retval = __kill(P(0));
					break;
				case _mesver:
					retval = 0x020311;
					break;
				case _usb_setup:
					retval = usb_setup(P(0), (char*)P(1), (char*)P(2));
					break;
				case _usb_write:
					retval = usb_write(P(0), (char*)P(1), P(2), P(3));
					break;
				case _usb_read:
					retval = usb_read(P(0), (char*)P(1), P(2), P(3));
					break;
				case _tcp_status:
					retval = tcp_status(id, P(0));
					break;
				case _fat_status:
					retval = fat_status(P(0), (int*)P(1), (int*)P(2), (char*)P(3));
					break;
				case _ret_value:
					retval = tcbptr->ret_value;
					break;
				case _set_budwidth:
					set_buswidth(P(0), P(1), P(2));
					break;
				case _add_device:
					retval = add_device(id, (void*)P(0));
					break;
				case _ddr_set:
					ddr_set(P(0), P(1));
					break;
				case _ddr_reset:
					ddr_reset(P(0), P(1));
					break;
				case _set_regist:
					retval = set_regist(P(0), (char*)P(1), (char*)P(2));
					break;
				case _get_regist:
					retval = get_regist(P(0), (char*)P(1), (char*)P(2), P(3));
					break;
				case _list_regist:
					retval = list_regist(P(0), P(1), (char*)P(2));
					break;
				case _hw_config:
					retval = hw_config(P(0), P(1), P(2));
					break;
				case _add_device_multi:
					retval = add_device_multi(id, (void*)P(0));
					break;
				case _romfile:
					retval = (int)get_romconfig((int*)P(0), (char*)P(1));
					break;
				}
				if(complete) {
					tcbptr->registers[0] = retval;
					reset_task_flag(id, TASKWAIT);
				}
			}
		}
	}
}
